from abc import ABC, abstractmethod
from mobikit_utils.utils import raise_for_keys, raise_for_values, raise_for_types
from mobikit_utils.query_builder.constants import (
    filter_types,
    payload_schema,
    sort_dirs,
    aggregation_types,
)
from mobikit.feeds.exceptions import (
    InvalidFilter,
    MalformedFilter,
    InvalidColumn,
    InvalidComparand,
    InvalidLowerBound,
    InvalidUpperBound,
    InvalidCoordinate,
    MalformedCoordinate,
    InvalidLongitude,
    InvalidLatitude,
    InvalidRadius,
    MalformedAggregator,
    MalformedSelector,
    InvalidGeometry,
)


class Query:
    """
    Top-level class for constructing query API paylaods.

    Parameters
    ----------
    select : list
        List of columns to return. List members may be instances of either Column or Aggregator.
    sample : float
        A sample percentage expressed as a value between 0.0 and 1.0.
    filterby : Filter or AtomicFilter
        The filter(s) that will be applied to the query.
    groupby : list
        List of columns to group by. List members must be instances of Column.
    sortby: list
        List of columns to sort by. List members must be instances of SortDirection.
    limit : int
        Maximum number of rows to return.
    """

    schema = payload_schema

    def __init__(
        self,
        select=None,
        filterby=None,
        groupby=None,
        sortby=None,
        limit=None,
        sample=None,
    ):
        select = select or []
        filterby = filterby or Filter()
        groupby = groupby or []
        sortby = sortby or []
        self._select = Select(*select)
        self._sample = Sample(sample) if sample is not None else sample
        self._filter = filterby
        self._group = Group(*groupby)
        self._sort = Sort(*sortby)
        self._limit = Limit(limit) if limit is not None else limit

    def to_dict(self):
        data = {
            self.schema.select.name: self._select,
            self.schema.filter.name: self._filter.to_dict(),
            self.schema.group.name: self._group,
            self.schema.sort.name: self._sort,
            self.schema.meta.name: {self.schema.meta.spatial_format: "text"},
        }
        if self._sample is not None:
            data[self.schema.sample.name] = self._sample.val
        if self._limit is not None:
            data[self.schema.limit.name] = self._limit.val
        return data


class Sample:
    """
    validates and constructs a well-formed sample portion of the query payload
    'val' property returns the serializable value
    """

    @classmethod
    def validate(cls, sample):
        tmp_key = "sample"
        raise_for_types(
            {tmp_key: sample},
            tmp_key,
            allowed_types=[float],
            ExceptionClass=InvalidColumn,
        )
        if (sample < 0) or (sample > 1):
            raise InvalidColumn(
                f"Sample must be a value between 0.0 and 1.0. got {sample}."
            )
        return sample

    @property
    def val(self):
        return self._val

    def __init__(self, sample):
        self._val = self.validate(sample)


class Limit:
    """
    validates and constructs a well-formed limit portion of the query payload
    'val' property returns the serializable value
    """

    @classmethod
    def validate(cls, limit):
        tmp_key = "limit"
        raise_for_types(
            {tmp_key: limit}, tmp_key, allowed_types=[int], ExceptionClass=InvalidColumn
        )
        return limit

    @property
    def val(self):
        return self._val

    def __init__(self, limit):
        self._val = self.validate(limit)


class Select(list):
    """
    validates and constructs a well-formed select portion of the query payload
    """

    @classmethod
    def validate(cls, selections):
        tmp_key = "select_element"
        for selection in selections:
            raise_for_types(
                {tmp_key: selection},
                tmp_key,
                allowed_types=[Column, Selector],
                ExceptionClass=InvalidColumn,
            )
        return selections

    def __init__(self, *selections):
        selections = self.validate(selections)
        super().__init__(selections)


class Group(list):
    """
    validates and constructs a well-formed group portion of the query payload
    """

    @classmethod
    def validate(cls, groupings):
        tmp_key = "groupby_element"
        for grouping in groupings:
            raise_for_types(
                {tmp_key: grouping},
                tmp_key,
                allowed_types=[Column],
                ExceptionClass=InvalidColumn,
            )
        return groupings

    def __init__(self, *groupings):
        groupings = self.validate(groupings)
        super().__init__(groupings)


class Sort(list):
    """
    validates and constructs a well-formed sort portion of the query payload
    """

    @classmethod
    def validate(cls, sorts):
        tmp_key = "sort_element"
        for sort in sorts:
            raise_for_types(
                {tmp_key: sort},
                tmp_key,
                allowed_types=[SortDirection],
                ExceptionClass=InvalidColumn,
            )
        return sorts

    def __init__(self, *sorts):
        sorts = self.validate(sorts)
        super().__init__(sorts)


class Filter(dict):
    """
    Validates and constructs a well-formed filter portion of the query payload.

    Parameters
    ----------
    *args Filter or AtomicFilter
        Any number of predicate Filters from which to compose this Filter. 
        Predicate filters must be instances of Filter or AtomicFilter.
    conjunction : {'and', 'or'}
        The logical conjunction that will be used to join `args`.    
    """

    schema = payload_schema.filter_block

    def __init__(self, *args, conjunction="and"):
        for arg in args:
            if not isinstance(arg, Filter) and not isinstance(arg, AtomicFilter):
                raise InvalidFilter(
                    f"Filter args must be of type Filter or AtomicFilter. Got {type(arg)} instead."
                )
        self._predicates = list(args)
        self._conjunction = conjunction

    def to_dict(self):
        return {
            self.schema.conjunction: self._conjunction,
            self.schema.predicates: [
                predicate.to_dict() for predicate in self._predicates
            ],
        }

    def __bool__(self):
        return True


class AtomicFilter(dict):
    """
    automatically generated from Column functions. no need to construct explicitly
    """

    schema = payload_schema.filter

    @classmethod
    def validate(self, data):
        data = raise_for_keys(
            data,
            [self.schema.field, self.schema.type, self.schema.meta],
            ExceptionClass=MalformedFilter,
        )
        data[self.schema.negated] = data.get(self.schema.negated, False)
        tmp_key = "negated"
        raise_for_types(
            {tmp_key: data[self.schema.negated]},
            tmp_key,
            allowed_types=[bool],
            ExceptionClass=MalformedFilter,
        )
        return data

    def negate(self):
        self[self.schema.negated] = (
            not self[self.schema.negated] if self[self.schema.negated] else True
        )

    def to_dict(self):
        return self

    def __bool__(self):
        return True

    def __neg__(self):
        self.negate()
        return self

    def __init__(self, data):
        data = self.validate(data)
        super().__init__(**data)


class Selectable(ABC):
    schema = payload_schema.select

    @abstractmethod
    def alias(self, name):
        """
        Creates an alias for the Column or Selector.

        Parameters
        ----------
        name : str
            Name to be used as the alias

        Returns
        ----------
        Selector
            Aliased Selector.
        """
        pass

    @abstractmethod
    def _set_format(self, val):
        """private method for spatial formatting operations"""
        pass

    def as_geojson(self):
        """
        Formats the spatial Column or Selector as GeoJSON.

        Returns
        ----------
        Selector
            Formatted Selector.
        """
        return self._set_format("geojson")

    def as_lat(self):
        """
        Formats the spatial Column or Selector as a numeric latitude.

        Returns
        ----------
        Selector
            Formatted Selector.
        """
        return self._set_format("lat")

    def as_lng(self):
        """
        Formats the spatial Column or Selector as a numeric longitude.

        Returns
        ----------
        Selector
            Formatted Selector.
        """
        return self._set_format("lng")

    def as_text(self):
        """
        Formats the spatial Column or Selector as well-known text.

        Returns
        ----------
        Selector
            Formatted Selector.
        """
        return self._set_format("text")


class Selector(dict, Selectable):
    """
    represents a validated and well-formed selector dict for a query payload
    automatically generated from Column functions. no need to construct explicitly
    """

    schema = payload_schema.select

    @classmethod
    def validate(cls, data):
        data = raise_for_keys(
            data, [cls.schema.field], ExceptionClass=MalformedSelector
        )
        return data

    def __init__(self, data):
        data = self.validate(data)
        super().__init__(**data)

    def alias(self, val):
        self[self.schema.alias] = val
        return self

    def _set_format(self, val):
        self[self.schema.format] = val
        return self


class Aggregator(Selector):
    """
    represents a validated and well-formed aggregator dict for a query payload
    automatically generated from Column functions. no need to construct explicitly
    """

    schema = payload_schema.select

    @classmethod
    def validate(self, data):
        data = raise_for_keys(
            data, [self.schema.function], ExceptionClass=MalformedAggregator
        )
        return data

    def __init__(self, data):
        data = self.validate(data)
        super().__init__(data)


class SortDirection(dict):
    """
    represents a validated and well-formed sort-direction dict for a query payload
    automatically generated from Column functions. no need to construct explicitly
    """

    schema = payload_schema.sort

    @classmethod
    def validate(self, data):
        data = raise_for_keys(data, [self.schema.field, self.schema.dir])
        return data

    def __init__(self, data):
        data = self.validate(data)
        super().__init__(**data)


class Column(str, Selectable):
    """
    A column identifier.

    Contains functions to construct atomic filters, aggregators, sort directions

    Parameters
    ----------
    identifier : str
        Name of the desired column.
    """

    schema = payload_schema.filter

    def __init__(self, identifier):
        self._identifier = self.validate(identifier)

    def alias(self, val):
        return Selector(
            {
                payload_schema.select.field: self._identifier,
                payload_schema.select.alias: val,
            }
        )

    def _set_format(self, val):
        return Selector(
            {
                payload_schema.select.field: self._identifier,
                payload_schema.select.format: val,
            }
        )

    def ascending(self):
        """
        Creates an ascending sort direction for the Column.

        Returns
        ----------
        SortDirection
            Created ascending sort direction.
        """
        return self._compose_sort_direction(sort_dirs.ascending)

    def descending(self):
        """
        Creates a descending sort direction for the Column.

        Returns
        ----------
        SortDirection
            Created descending sort direction.
        """
        return self._compose_sort_direction(sort_dirs.descending)

    def stddev(self):
        """
        Creates a stddev aggregator for the column.

        Returns
        ----------
        Aggregator
            Created stddev aggregator.
        """
        return self._compose_aggregator(aggregation_types.stddev)

    def median(self):
        """
        Creates a median aggregator for the column.

        Returns
        ----------
        Aggregator
            Created median aggregator.
        """
        return self._compose_aggregator(aggregation_types.median)

    def min(self, distinct=False):
        """
        Creates a min aggregator for the column.
        
        Parameters
        ----------
        distinct : bool
            If true, will only consider unique values in the aggregation.

        Returns
        -------
        Aggregator
            Created min aggregator.
        """
        return self._compose_aggregator(aggregation_types.min, distinct=distinct)

    def max(self, distinct=False):
        """
        Creates a max aggregator for the column.
        
        Parameters
        ----------
        distinct : bool
            If true, will only consider unique values in the aggregation.

        Returns
        -------
        Aggregator
            Created max aggregator.
        """
        return self._compose_aggregator(aggregation_types.max, distinct=distinct)

    def first(self, distinct=False):
        """
        Creates a first aggregator for the column.
        
        Parameters
        ----------
        distinct : bool
            If true, will only consider unique values in the aggregation.

        Returns
        -------
        Aggregator
            Created first aggregator.
        """
        return self._compose_aggregator(aggregation_types.first, distinct=distinct)

    def last(self, distinct=False):
        """
        Creates a last aggregator for the column.
        
        Parameters
        ----------
        distinct : bool
            If true, will only consider unique values in the aggregation.

        Returns
        -------
        Aggregator
            Created last aggregator.
        """
        return self._compose_aggregator(aggregation_types.last, distinct=distinct)

    def count(self, distinct=False):
        """
        Creates a count aggregator for the column.
        
        Parameters
        ----------
        distinct : bool
            If true, will only consider unique values in the aggregation.

        Returns
        -------
        Aggregator
            Created count aggregator.
        """
        return self._compose_aggregator(aggregation_types.count, distinct=distinct)

    def sum(self, distinct=False):
        """
        Creates a sum aggregator for the column.
        
        Parameters
        ----------
        distinct : bool
            If true, will only consider unique values in the aggregation.

        Returns
        -------
        Aggregator
            Created sum aggregator.
        """
        return self._compose_aggregator(aggregation_types.sum, distinct=distinct)

    def average(self, distinct=False):
        """
        Creates an average aggregator for the column.
        
        Parameters
        ----------
        distinct : bool
            If true, will only consider unique values in the aggregation.

        Returns
        -------
        Aggregator
            Created average aggregator.
        """
        return self._compose_aggregator(aggregation_types.average, distinct=distinct)

    def equals(self, comparand, negated=False):
        """
        Creates an equals AtomicFilter.
        
        Parameters
        ----------
        comparand : Column or str or int or float or ISO-8601 datestring
            The value to which the column will be compared.
        negated : bool
            If true, resulting AtomicFilter will be negated.

        Returns
        -------
        AtomicFilter
            Created equals AtomicFilter.
        """
        return self._compose_ineqality_filter(
            filter_types.equal, comparand, negated=negated
        )

    def greater_than(self, comparand, negated=False):
        """
        Creates a greater than AtomicFilter.
        
        Parameters
        ----------
        comparand : Column or str or int or float or ISO-8601 datestring
            The value to which the column will be compared.
        negated : bool
            If true, resulting AtomicFilter will be negated.

        Returns
        -------
        AtomicFilter
            Created greater than AtomicFilter.
        """
        return self._compose_ineqality_filter(
            filter_types.greater_than, comparand, negated=negated
        )

    def less_than(self, comparand, negated=False):
        """
        Creates a less than AtomicFilter.
        
        Parameters
        ----------
        comparand : Column or str or int or float or ISO-8601 datestring
            The value to which the column will be compared.
        negated : bool
            If true, resulting AtomicFilter will be negated.

        Returns
        -------
        AtomicFilter
            Created less than AtomicFilter.
        """
        return self._compose_ineqality_filter(
            filter_types.less_than, comparand, negated=negated
        )

    def greater_than_equal(self, comparand, negated=False):
        """
        Creates a greater than or equal to AtomicFilter.
        
        Parameters
        ----------
        comparand : Column or str or int or float or ISO-8601 datestring
            The value to which the column will be compared.
        negated : bool
            If true, resulting AtomicFilter will be negated.

        Returns
        -------
        AtomicFilter
            Created greater than or equal to AtomicFilter.
        """
        return self._compose_ineqality_filter(
            filter_types.greater_than_equal, comparand, negated=negated
        )

    def less_than_equal(self, comparand, negated=False):
        """
        Creates a less than or equal to AtomicFilter.
        
        Parameters
        ----------
        comparand : Column or str or int or float or ISO-8601 datestring
            The value to which the column will be compared.
        negated : bool
            If true, resulting AtomicFilter will be negated.

        Returns
        -------
        AtomicFilter
            Created less than or equal to AtomicFilter.
        """
        return self._compose_ineqality_filter(
            filter_types.less_than_equal, comparand, negated=negated
        )

    def isin(self, *comparand, negated=False):
        """
        Creates an 'in' AtomicFilter.
        
        Parameters
        ----------
        *comparand : Column or str or int or float or ISO-8601 datestring
            Any number of comparands that will form the set used to filter results.
        negated : bool
            If true, resulting AtomicFilter will be negated.

        Returns
        -------
        AtomicFilter
            Created 'in' AtomicFilter.
        """
        comparand = list(comparand)
        for val in comparand:
            tmp_key = "comparand_element"
            raise_for_types(
                {tmp_key: val},
                tmp_key,
                allowed_types=[str, int, float],
                ExceptionClass=InvalidComparand,
            )
        data = {
            self.schema.field: self._identifier,
            self.schema.type: filter_types.is_in,
            self.schema.negated: negated,
            self.schema.meta: {
                self.schema.in_meta.comparand: self._to_argument(comparand)
            },
        }
        return AtomicFilter(data)

    def like(self, comparand, negated=False):
        """
        Creates an like AtomicFilter.
        
        Parameters
        ----------
        comparand : str
            A patten string, to be used with a SQL 'LIKE' operator.
        negated : bool
            If true, resulting AtomicFilter will be negated.

        Returns
        -------
        AtomicFilter
            Created like AtomicFilter.
        """
        tmp_key = "comparand"
        raise_for_types(
            {tmp_key: comparand},
            tmp_key,
            allowed_types=[str],
            ExceptionClass=InvalidComparand,
        )
        data = {
            self.schema.field: self._identifier,
            self.schema.type: filter_types.like,
            self.schema.negated: negated,
            self.schema.meta: {
                self.schema.like_meta.comparand: self._to_argument(comparand)
            },
        }
        return AtomicFilter(data)

    def between(self, lb, ub, negated=False):
        """
        Creates a between/range AtomicFilter.
        
        Parameters
        ----------
        lb : Column or str or int or float or ISO-8601 datestring
            Value to be used as the lower bound (inclusive).
        ub : Column or str or int or float or ISO-8601 datestring
            Value to be used as the upper bound (inclusive).
        negated : bool
            If true, resulting AtomicFilter will be negated.

        Returns
        -------
        AtomicFilter
            Created between/range AtomicFilter.
        """
        tmp_key = ("lb", "ub")
        raise_for_types(
            {tmp_key[0]: lb},
            tmp_key[0],
            allowed_types=[Column, int, float, str],
            ExceptionClass=InvalidLowerBound,
        )
        raise_for_types(
            {tmp_key[1]: ub},
            tmp_key[1],
            allowed_types=[Column, int, float, str],
            ExceptionClass=InvalidUpperBound,
        )
        data = {
            self.schema.field: self._identifier,
            self.schema.type: filter_types.between,
            self.schema.negated: negated,
            self.schema.meta: {
                self.schema.between_meta.lb: self._to_argument(lb),
                self.schema.between_meta.ub: self._to_argument(ub),
            },
        }
        return AtomicFilter(data)

    def _compose_aggregator(self, aggregation_type, distinct=False):
        schema = payload_schema.select
        data = {
            schema.field: self._identifier,
            schema.function: aggregation_type,
            schema.distinct: distinct,
        }
        return Aggregator(data)

    def _compose_sort_direction(self, sort_dir):
        schema = payload_schema.sort
        data = {schema.field: self._identifier, schema.dir: sort_dir}
        return SortDirection(data)

    def _compose_ineqality_filter(self, filter_type, comparand, negated=False):
        tmp_key = "comparand"
        raise_for_types(
            {tmp_key: comparand},
            tmp_key,
            allowed_types=[Column, str, int, float],
            ExceptionClass=InvalidComparand,
        )
        data = {
            self.schema.field: self._identifier,
            self.schema.type: filter_type,
            self.schema.negated: negated,
            self.schema.meta: {
                self.schema.inequality_meta.comparand: self._to_argument(comparand)
            },
        }
        return AtomicFilter(data)

    @classmethod
    def _to_argument(self, val):
        return "${" + val + "}" if isinstance(val, Column) else val

    @classmethod
    def validate(self, identifier):
        tmp_key = "column_identifier"
        raise_for_types(
            {tmp_key: identifier},
            tmp_key,
            allowed_types=[str],
            ExceptionClass=InvalidColumn,
        )
        return identifier

    def __lt__(self, o):
        return self.less_than(o)

    def __gt__(self, o):
        return self.greater_than(o)

    def __le__(self, o):
        return self.less_than_equal(o)

    def __ge__(self, o):
        return self.greater_than_equal(o)

    def __eq__(self, o):
        return self.equals(o)

    def __ne__(self, o):
        return self.equals(o, negated=True)

    def __isin__(self, o):
        return self.isin(*o)

    def __repr__(self):
        return self._identifier


class SpatialColumn(Column):
    """
    A spatial column identifier.

    Contains additional methods to construct spatial AtomicFilters.

    Parameters
    ----------
    identifier : str
        Name of the desired column.
    """

    def within(self, point, radius, negated=False):
        """
        Creates a within AtomicFilter for a spatial column.

        Parameters
        ----------
        point : tuple of the form (lat, lng)
            Center point of the within filter. `lat` and `lng` should both be valid lat/lng floats.
        radius : int or float
            Distance in meters from the center point.
        negated : bool
            If true, resulting AtomicFilter will be negated.

        Returns
        -------
        AtomicFilter
            Created within AtomicFilter.
        """
        lat, lng = self._validate_point(point)
        tmp_key = "radius"
        raise_for_types(
            {tmp_key: radius},
            tmp_key,
            allowed_types=[int, float],
            ExceptionClass=InvalidComparand,
        )
        data = {
            self.schema.field: self._identifier,
            self.schema.type: filter_types.within,
            self.schema.negated: negated,
            self.schema.meta: {
                self.schema.within_meta.lat: lat,
                self.schema.within_meta.lng: lng,
                self.schema.within_meta.radius: radius,
            },
        }
        return AtomicFilter(data)

    def intersected_by(self, geometry, negated=False):
        """
        Creates an intersected by (point in polygon) AtomicFilter for a spatial column.

        Parameters
        ----------
        geometry : Column or list of points.
            Each point should be a tuple of the form (lat, lng). 
            `lat` and `lng` should both be valid lat/lng floats.
        negated : bool
            If true, resulting AtomicFilter will be negated.

        Returns
        -------
        AtomicFilter
            Created intersected by AtomicFilter.
        """
        tmp_key = "geometry"
        raise_for_types(
            {tmp_key: geometry},
            tmp_key,
            allowed_types=[list, SpatialColumn],
            ExceptionClass=InvalidGeometry,
        )
        if isinstance(geometry, list):
            for point in geometry:
                self._validate_point(point)
            geometry = [
                list(point) if isinstance(point, tuple) else point for point in geometry
            ]

        data = {
            self.schema.field: self._identifier,
            self.schema.type: filter_types.intersected_by,
            self.schema.negated: negated,
            self.schema.meta: {
                self.schema.intersected_by_meta.geometry: self._to_argument(geometry)
            },
        }
        return AtomicFilter(data)

    def _validate_point(self, point):
        tmp_key = "coordinate"
        raise_for_types(
            {tmp_key: point},
            tmp_key,
            allowed_types=[tuple, list],
            ExceptionClass=InvalidCoordinate,
        )
        if len(point) != 2:
            raise MalformedCoordinate(
                f"Expected a coordinate of the form (lat, lng), instead got {point}"
            )
        lat, lng = point
        tmp_key = ("lat", "lng")
        raise_for_types(
            {tmp_key[0]: lat},
            tmp_key[0],
            allowed_types=[int, float],
            ExceptionClass=InvalidComparand,
        )
        raise_for_types(
            {tmp_key[1]: lng},
            tmp_key[1],
            allowed_types=[int, float],
            ExceptionClass=InvalidComparand,
        )
        return lat, lng



from functools import partial


class Infix(object):
    def __init__(self, func):
        self._func = func

    def __ror__(self, o):
        return Infix(partial(self._func, o))

    def __or__(self, o):
        return self._func(o)

    def __call__(self, o1, o2):
        return self._func(o1, o2)


@Infix
def isin(o1, o2):
    return o1.__isin__(o2)



from mobikit.exceptions import MobikitException


class SourceTypeException(MobikitException):
    """raised when trying to query a non-queryable source type"""


class PayloadException(MobikitException):
    """used to raise exception for invalid query payload"""


class SourceError(MobikitException):
    """used to raise exceptions for missing and/or type-mismatched feed sources"""


class FeedError(MobikitException):
    """used to raise exceptions when loading sources for a feed"""


class InvalidFilter(MobikitException):
    """raised when an arg for Filter is of an invalid type"""


class MalformedFilter(MobikitException):
    """raised when an atomic filter is improperly formatted"""


class InvalidColumn(MobikitException):
    """raised when a column is constructed with invalid args"""


class InvalidComparand(MobikitException):
    """raised when a comparand provided to a filter constructing function is invalid"""


class InvalidLowerBound(MobikitException):
    """raised when an lower bound provided to a filter constructing function is invalid"""


class InvalidUpperBound(MobikitException):
    """raised when an upper bound provided to a filter constructing function is invalid"""


class InvalidCoordinate(MobikitException):
    """raised when the coordinate provided to a filter constructing function is invalid"""


class MalformedCoordinate(MobikitException):
    """raised when either the latitude or longitude portion of a coordinate are malformed"""


class InvalidLongitude(MobikitException):
    """raised when a longitude provided to a filter constructing function is invalid"""


class InvalidLatitude(MobikitException):
    """raised when a latitude provided to a filter constructing function is invalid"""


class InvalidRadius(MobikitException):
    """raised when a radius provided to a filter constructing function is invalid"""


class MalformedAggregator(MobikitException):
    """raised when an aggregator provides malformed data"""


class MalformedSelector(MobikitException):
    """raised when an selector provides malformed data"""


class InvalidGeometry(MobikitException):
    """raised when a provided gerometry is improperly formatted"""


class DBSourceError(MobikitException):
    """raised when an error is encountered when loading/querying a db source"""


class FileSourceError(MobikitException):
    """raised when an error is encountered when loading/querying a file source"""



import pandas as pd
import json
import os
import requests
import logging
from mobikit.config import config
from mobikit.exceptions import AuthException
from mobikit.utils import token_required
from mobikit.feeds.base import FeedRef
from mobikit.exceptions import MobikitException


@token_required
def add_data_source(dataframe, feed_id, source_name):
    """
    Adds a source to the datafeed from a local datasource. Must be a
    pandas DataFrame.

    Parameters
    ----------
    dataframe : pandas.DataFrame
        DataFrame containing data
    feed_id : str, int
        String containing the feed id
    source_name : str
        Name of source

    Returns
    -------
    None : None
    """
    logging.warn("This function is temporarily unavailable.")
    return

    headers = {"Authorization": "Token {}".format(config.api_token)}
    try:
        if not (isinstance(dataframe, pd.DataFrame) and _validate_feed_id(feed_id)):
            raise MobikitException("please pass in a dataframe and a valid feed_id")

        if dataframe.memory_usage(index=True).sum() >= config.upload_size_limit:
            raise MobikitException("dataframe too big! must be <250mb")

        # upload resource to s3
        resource_url = _upload_file(headers, dataframe)

        # add source to feed
        _add_source_to_feed(headers, feed_id, resource_url, source_name)

        # check that source was added correctly
        _validate_source(headers, "derived_data", resource_url)
        config.logger.info("success! added source to datafeed: " + str(feed_id))

    except Exception as e:
        config.logger.exception(e)
        raise MobikitException("Error adding data source.") from e


# method to confirm that the feed id passed is valid
def _validate_feed_id(feed_id):
    FeedRef.create(feed_id).load(meta_only=True)
    return True


def _upload_file(headers, dataframe):
    try:
        os.makedirs("cache/", exist_ok=True)
        dataframe.to_csv("cache/dummy.csv", index=False)
        with open("cache/dummy.csv", "rb") as f:
            response = requests.post(
                config.upload_route, headers=headers, files=[("file", f)]
            )
            response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        config.logger.exception(e)
    except requests.exceptions.RequestException as e:
        config.logger.exception(e)
    except Exception as e:
        config.logger.exception(e)
    finally:
        os.remove("cache/dummy.csv")
    return response.json()["resource_url"]


def _add_source_to_feed(headers, feed_id, resource_url, source_name):
    try:
        payload = {
            "resource": resource_url,
            "cadence": "never",
            "source_type": "derived_data",
            "dataset_id": feed_id,
            "name": source_name,
        }
        resource = requests.post(config.source_route, headers=headers, json=payload)
        resource.raise_for_status()
        return resource.content
    except requests.exceptions.HTTPError as e:
        config.logger.exception(e)
    except requests.exceptions.RequestException as err:
        config.logger.exception(err)
    except Exception as e:
        config.logger.exception(e)


# method to validate that the source was successfully created
def _validate_source(headers, resource_type, resource_url):
    try:
        payload = {"type": resource_type, "resource": resource_url}
        resource = requests.post(config.validate_route, headers=headers, data=payload)
        resource.raise_for_status()
    except requests.exceptions.HTTPError as e:
        config.logger.exception(e)
    except requests.exceptions.RequestException as err:
        config.logger.exception(err)
    except Exception as e:
        config.logger.exception(e)



import os
import pkg_resources
import json
import logging
import mobikit.feeds as feeds
import mobikit.feeds as workspaces
from .config import config

# this will intentionally error for public release
# we don't want to expose models package to the public
try:
    from .models import models
except ImportError:
    pass

# instantiate logger
config.logger = logging.getLogger()
console = logging.StreamHandler()
config.logger.addHandler(console)
# check the environoment and set api base
host = os.getenv("MOBIKIT_API_HOST")
if not host:
    config.base = "https://api.mobikit.io/"
elif "http" in host:
    config.base = host
elif "localhost" in host:
    config.base = f"http://{host}/"
else:
    config.base = f"https://api.{host}/"

# load api urls based no environment
constants_file = pkg_resources.resource_filename("mobikit", "reference/constants.json")
with open(constants_file) as f:
    constants = json.load(f)
    config.upload_route = config.base + constants["urls"]["upload_datafeed_route"]
    config.validate_route = config.base + constants["urls"]["validate"]
    config.feed_route = config.base + constants["urls"]["feeds"]
    config.source_route = config.base + constants["urls"]["sources"]
    config.query_route = config.base + constants["urls"]["query"]


# allow user to pass in an api key to get started
def set_api_key(user_token):
    if not isinstance(user_token, str):
        config.logger.exception("pass in a string token!")
    else:
        config.api_token = user_token



class MobikitException(Exception):
    """generic exception raised from within the mobikit library"""


class AuthException(MobikitException):
    def __init__(self):
        self.message = "You have not passed in a authorization token. Try using mobikit.set_api_key(<your token string>)"
        Exception.__init__(self, self.message)



import json
import requests
from mobikit.config import config
from mobikit.utils import token_required
from mobikit.exceptions import AuthException, MobikitException
from mobikit.feeds.base import FeedRef


@token_required
def search(keywords=None):
    """
    Function to allow a user to search for their workspaces with keywords.

    Parameters
    ----------
    keywords : str
        String of keywords

    Returns
    -------
    search_results : list<object>
    a list of workspaces
    """
    if not keywords:
        keywords = ""
    if not isinstance(keywords, str):
        raise MobikitException("pass in a string for keywords")
    headers = {"Authorization": "Token " + config.api_token}
    query_string = {"query": keywords, "page_size": 100}
    try:
        f = requests.get(config.feed_route, headers=headers, params=query_string)
        f.raise_for_status()
        search_results = []
        for feed_result in json.loads(f.content):
            feed = FeedRef.create(feed_result["id"])
            feed.load(meta_only=True)
            search_results.append(
                {
                    "workspace_id": feed.id,
                    "workspace_name": feed.name,
                    "feeds": list(feed.sources.keys()),
                }
            )
        config.logger.info(
            "your search yielded the following feeds: " + str(search_results)
        )
        return search_results
    except requests.exceptions.HTTPError as e:
        config.logger.exception(e)
        response_status = e.response.content if e.response is not None else "unknown"
        raise MobikitException(
            f"Recieved invalid response status code {response_status}."
        ) from e
    except requests.exceptions.RequestException as e:
        config.logger.exception(e)
        raise MobikitException(f"Error retrieving search results.") from e



from mobikit.config import config
from mobikit.exceptions import AuthException
from functools import wraps


def token_required(func):
    @wraps(func)
    def _token_required(*args, **kwargs):
        if config.api_token is None:
            raise AuthException
        return func(*args, **kwargs)

    return _token_required



api_token = None
base = None
query_route = None
feed_route = None
source_route = None
upload_route = None
validate_route = None
file_size_limit = 250000000
upload_size_limit = 20000000
chunk_size = 200000
logger = None
col_ref = None
counties_ref = None
meta_base = None



import pandas as pd
import json
from json import JSONDecodeError
import requests
import io
from abc import ABC, abstractmethod
from mobikit_utils.utils import raise_for_types
from mobikit_utils.query_builder.constants import payload_schema, spatial_formats
from mobikit.exceptions import MobikitException
from mobikit.config import config
from mobikit.feeds.exceptions import (
    FeedError,
    SourceError,
    DBSourceError,
    FileSourceError,
)
from requests.exceptions import RequestException, HTTPError
from mobikit.feeds.exceptions import PayloadException
from mobikit.feeds.load.query_builder import Query


class Frameable(ABC):
    @abstractmethod
    def to_df(self):
        """returns a df representation of loaded data"""
        pass


class Loadable(ABC):
    @abstractmethod
    def load(self):
        """loads appropriate source and returns self (for chaining)"""
        pass

    @abstractmethod
    def _handle_request_exception(e, ExceptionClass=MobikitException):
        """gets called whenever a request-realated exception is caught"""
        pass

    def __init__(self, *args, **kwargs):
        self._headers = {"Authorization": "Token {}".format(config.api_token)}
        self._params = {"presigned": "localhost" not in config.base}


class Queryable(Loadable, ABC):
    @abstractmethod
    def query(self, payload):
        """Abstract method to execute db query with provided payload"""
        pass

    def load(self, *args, **kwargs):
        return self.query(*args, **kwargs)

    @classmethod
    def _validate_payload(cls, payload):
        payload = payload or {}
        tmp_key = "query_payload"
        raise_for_types(
            {tmp_key: payload},
            tmp_key,
            allowed_types=[str, dict, Query],
            ExceptionClass=PayloadException,
        )
        if isinstance(payload, str):
            try:
                payload = json.loads(payload)
            except JSONDecodeError as e:
                raise PayloadException("Query payload cannot be parsed as valid JSON")
        elif isinstance(payload, dict):  # type(payload) == dict
            try:
                json.dumps(payload)
            except TypeError as e:
                raise PayloadException("Query payload cannot be parsed as valid JSON")
        else:  # payload is Query
            payload = payload.to_dict()
        payload[payload_schema.meta.name] = {
            payload_schema.meta.spatial_format: spatial_formats.text
        }
        return payload


class BaseSourceRef(Loadable, Frameable, ABC):
    """A reference to a particular data source"""

    def __init__(self, feed_id, source):
        super().__init__(feed_id, source)
        self._feed_id = feed_id
        self._id = source.get("id", None)
        self._name = source.get("name", None)
        self._source = source
        self._view = pd.DataFrame()

    @property
    def name(self):
        return self._source["name"]

    @classmethod
    def create(cls, feed_id, source, **kwargs):
        cls.validate(feed_id, source, **kwargs)
        return cls(feed_id, source, **kwargs)

    @classmethod
    def validate(cls, feed_id, source, **kwargs):
        if not isinstance(feed_id, int):
            raise SourceError(
                f"'feed_id' must be of type <int>. got {type(feed_id)} instead"
            )
        if not isinstance(source, dict):
            raise SourceError(
                f"'source_name' must be of type <str>. got {type(dataset_source_id)} instead"
            )
        # TODO: check for id, resource keys on source, url key on resource

    def _handle_request_exception(self, e, ExceptionClass=MobikitException):
        if isinstance(e, HTTPError):
            res_status = e.response.status_code if e.response is not None else "unknown"
            raise ExceptionClass(
                f"Invalid response status code '{res_status}' recieved when loading source '{self._name}'."
            ) from e
        else:
            raise ExceptionClass(f"Error loading source '{self._name}'.") from e

    @classmethod
    @abstractmethod
    def _parse_response(cls, res):
        """responsible for parsing the raw response from API into view format"""
        pass

    def to_df(self):
        return self._view


class DBSourceRef(BaseSourceRef, Queryable):
    """A reference to a database table source"""

    def __init__(self, feed_id, source):
        super().__init__(feed_id, source)
        self._url = f"{config.query_route}{feed_id}/{self._id}/"

    def load(self):
        return self.query()

    def query(self, payload=None):
        payload = self._validate_payload(payload)

        try:
            res = requests.post(
                self._url, params=self._params, headers=self._headers, json=payload
            )
            res.raise_for_status()
        except RequestException as e:
            config.logger.exception(e)
            self._handle_request_exception(e, ExceptionClass=DBSourceError)
        self._view = self._parse_response(res)
        return self

    @classmethod
    def _parse_response(cls, res):
        res = res.json()
        if res.get("downsized", False):
            downsize_info = res.get("downsize_info", {})
            true_row_count = downsize_info.get("true_count", None)
            config.logger.warn(
                f"WARNING: Your request was very large ({true_row_count} rows), so the query result has been down sized. Please use the query field to narrow your request."
            )
        data = res.get("data")
        if isinstance(data, str):
            data = json.loads(data)
        return pd.DataFrame(data[1:], columns=data[0])


class FileSourceRef(BaseSourceRef):
    """A reference to an S3/CSV file source"""

    def __init__(self, feed_id, source):
        super().__init__(feed_id, source)
        self._url = self._source["resource"]["url"]

    def load(self):
        # try:
        #     res = requests.get(self._url)
        #     res.raise_for_status()
        # except Exception as e:
        #     config.logger.exception(e)
        #     self._handle_request_exception(e, ExceptionClass=FileSourceError)
        self._view = self._parse_response(self._url)
        return self

    @classmethod
    def _parse_response(cls, url):
        return pd.read_csv(url)
        # return pd.read_csv(io.StringIO(res.content.decode("utf-8")))


class FeedRef(Loadable):
    """A reference to a data feed"""

    def __init__(self, feed_id):
        super().__init__()
        self._id = feed_id
        self._sources = []
        self.name = None

    @property
    def id(self):
        return self._id

    @property
    def sources(self):
        return self._sources

    @classmethod
    def create(cls, feed_id):
        cls.validate(feed_id)
        return cls(feed_id)

    @classmethod
    def validate(cls, feed_id):
        if not isinstance(feed_id, int):
            raise FeedError(
                f"'feed_id' must be of type <int>. got {type(feed_id)} instead"
            )

    def _handle_request_exception(self, e, ExceptionClass=FeedError):
        if isinstance(e, HTTPError):
            res_status = e.response.status_code if e.response is not None else "unknown"
            raise ExceptionClass(
                f"Invalid response status code '{res_status}' recieved when loading feed '{self._id}'."
            ) from e
        else:
            raise ExceptionClass(f"Error loading feed '{self._id}'.") from e

    def get_source(self, source_name):
        source = self.sources.get(source_name, None)
        if source is None:
            raise FeedError(
                f"Feed with id: {self._id} does not contain a source with name: {source_name}."
            )
        return source

    def _ref_for_source(self, source, meta_only=False):
        if meta_only:
            return pd.DataFrame()
        if source["resource"]["resource_type"] == "db_table":
            return DBSourceRef.create(self._id, source)
        else:
            return FileSourceRef.create(self._id, source)

    def load(self, meta_only=False):
        try:
            res = requests.get(
                f"{config.feed_route}{self._id}/",
                params=self._params,
                headers=self._headers,
            )
            res.raise_for_status()
        except RequestException as e:
            config.logger.exception(e)
            self._handle_request_exception(e)
        data = json.loads(res.content)
        self.name = data["name"]
        self._sources = {
            source["name"]: self._ref_for_source(source, meta_only=meta_only)
            for source in data["sources"]
        }
        return self



from setuptools import setup
from types import SimpleNamespace
from setup_base import base_kwargs

setup(
    **base_kwargs,
    include_package_data=True,
    install_requires=[
        "mobikit-utils~=0.1.0",
        "bokeh>=1.0.0",
        "geopy>=1.19.0",
        "matplotlib>=3.0.3",
        "numpy>=1.16.1",
        "pandas>=0.24.2",
        "psycopg2-binary>=2.8.0",
        "requests>=2.21.0",
        "scikit-learn>=0.20.2",
        "scipy>=1.2.1",
        "geopy>=1.19.0",
        "plotly>=4.1.0",
        "scipy>=1.2.1",
        "shapely>=1.6.3",
        "geopandas>=0.5.0",
        "geopy>=1.19.0",
    ],
)



from setuptools import setup, find_packages
from setup_base import base_kwargs

setup(
    **base_kwargs,
    packages=find_packages(
        include=["mobikit", "mobikit.config", "mobikit.feeds", "mobikit.feeds.*"]
    ),
    zip_safe=False,
    include_package_data=True,
    install_requires=["pandas>=0.24.2", "requests>=2.21.0", "mobikit-utils~=0.1.0"],
)



from mobikit.config import config
from mobikit.feeds.exceptions import SourceTypeException
from mobikit.utils import token_required
from mobikit.feeds.base import FeedRef, Queryable


@token_required
def load(
    feed_id, source_name=None, query=None, source_names=None, queries=None, as_dict=True
):
    """
    Allows user to load/query any number of sources from a feed.

    If the request is very large, the result will be downsized and a warning will be issued to the user

    Parameters
    ----------
    feed_id : int
        Id of the feed whose sources you with to load. A feed consists of multiple sources.
    source_names : list of str
        List of source names to load data from.
    queries : dict
        Dict mapping source names (str) to queries (mobikit.feeds.Query).
    as_dict : boolean
        Whether the loaded dfs should be returned as dict or list.
    source_name : str
        String containing the name of a single source to query.
    query : mobikit.feeds.Query
        A query to be run on the single source specified with `source_name`.

    Returns
    -------
    dataframe(s) : pandas.DataFrame or dict or list
        If source_name is used to specify a single source to load, will return a single pandas.DataFrame.
        Otherwise, will return either a dict of the form {source_name: pandas.DataFrame}
        or a list of pandas.DataFrame, depending on the value of `as_dict`.
    """
    source_names = source_names or set()
    queries = queries or dict()
    if source_name:
        source_names.add(source_name)
        queries[source_name] = queries.get(source_name, None) or query

    feed = FeedRef.create(feed_id)
    feed.load()
    if not len(source_names):
        source_names = set(feed.sources.keys())

    source_dfs = {}
    for s in source_names:
        source_query = queries.get(s, None)
        if source_query and isinstance(feed.get_source(s), Queryable):
            feed.sources[s].query(source_query)
        elif source_query:
            raise SourceTypeException(f"{s} is not queryable.")
        else:
            feed.get_source(s).load()
        source_dfs[s] = feed.get_source(s).to_df()
    config.logger.info(f"returned {len(list(source_dfs.keys()))} dataframes")
    if source_name:
        return source_dfs[source_name]
    return source_dfs if as_dict else list(source_dfs.values())
